@@ -369,7 +369,6 @@ fn get_srp_client_public_bytes(client_public: &BigInt) -> Vec<u8> {
     v
 }
 
-#[derive(Debug)]
 pub struct WireProtocol {
     write_buf: Vec<u8>,
 
@@ -469,7 +468,7 @@ impl WireProtocol {
 
     fn send_packets(&mut self) -> Result<(), Error> {
         self.channel.write(&self.write_buf)?;
-        &self.write_buf.clear();
+        self.write_buf.clear();
         Ok(())
     }
 
@@ -566,6 +565,27 @@ impl WireProtocol {
         }
     }
 
+    fn guess_wire_crypt(&self, buf: &[u8]) -> (Vec<u8>, Vec<u8>) {
+        let mut params = HashMap::new();
+        let mut i: usize = 0;
+
+        while i < buf.len() {
+            let k = buf[i];
+            i += 1;
+            let ln = buf[i] as usize;
+            i += 1;
+            let v = &buf[i..i + ln];
+            i += ln;
+            params.insert(k, v);
+        }
+        if let Some(chacha) = params.get(&3) {
+            if &chacha[0..7] == b"ChaCha\x00" {
+                return (b"ChaCha".to_vec(), chacha[7..].to_vec());
+            }
+        }
+        return (b"Arc4".to_vec(), vec![]);
+    }
+
     pub fn parse_connect_response(
         &mut self,
         username: &str,
@@ -587,8 +607,8 @@ impl WireProtocol {
             panic!("connection error"); // not reach
         }
 
-        self.protocol_version = self.recv_packets(1)?[0] as i32;
         self.recv_packets(3)?;
+        self.protocol_version = self.recv_packets(1)?[0] as i32;
         self.accept_architecture = utils::bytes_to_buint32(&self.recv_packets(4)?) as i32;
         self.accept_type = utils::bytes_to_buint32(&self.recv_packets(4)?) as i32;
 
@@ -606,7 +626,7 @@ impl WireProtocol {
 
         // skip keys
         ln = utils::bytes_to_buint32(&self.recv_packets(4)?) as usize;
-        &self.recv_packets_alignment(ln);
+        self.recv_packets_alignment(ln)?;
 
         assert!(&self.accept_plugin_name == "Srp" || &self.accept_plugin_name == "Srp256");
 
@@ -621,14 +641,14 @@ impl WireProtocol {
             data = self.recv_packets_alignment(ln)?;
 
             ln = utils::bytes_to_buint32(&self.recv_packets_alignment(4)?) as usize;
-            &self.recv_packets_alignment(ln)?; // plugin_name
+            self.recv_packets_alignment(ln)?; // plugin_name
 
             ln = utils::bytes_to_buint32(&self.recv_packets_alignment(4)?) as usize;
-            &self.recv_packets_alignment(ln)?; // plugin_name_list
+            self.recv_packets_alignment(ln)?; // plugin_name_list
 
             // skip keys
             ln = utils::bytes_to_buint32(&self.recv_packets(4)?) as usize;
-            &self.recv_packets_alignment(ln);
+            self.recv_packets_alignment(ln)?;
         }
         ln = utils::bytes_to_uint16(&data[..2]) as usize;
         let servre_salt = &data[2..2 + ln];
@@ -644,14 +664,18 @@ impl WireProtocol {
             &self.accept_plugin_name,
         );
 
-        if opcode == OP_COND_ACCEPT {
+        let (encrypt_plugin, nonce) = if opcode == OP_COND_ACCEPT {
             self.op_cont_auth(&auth_data)?;
-            self.op_response()?;
-        }
+            let (_, _, buf) = self.op_response()?;
+            self.guess_wire_crypt(&buf)
+        } else {
+            (Vec::new(), Vec::new())
+        };
 
         if options["wire_crypt"] == "true" {
-            self.op_crypt()?;
-            self.channel.set_arc4_key(&session_key);
+            self.op_crypt(&encrypt_plugin)?;
+            self.channel
+                .set_crypt_key(&encrypt_plugin, &session_key, &nonce);
             self.op_response()?;
         } else {
             self.auth_data = Some(auth_data); // use in op_attach(), op_create()
@@ -832,12 +856,20 @@ impl WireProtocol {
         client_public: &BigInt,
     ) -> Result<(), Error> {
         debug_print!("op_connect()");
+        // PROTOCOL_VERSION, Arch type (Generic=1), min, max, weight
+        let protocols = [
+            "ffff800d00000001000000000000000500000008", // 13, 1, 0, 5, 8
+            "ffff800e0000000100000000000000050000000a", // 14, 1, 0, 5, 10
+            "ffff800f0000000100000000000000050000000c", // 15, 1, 0, 5, 12
+            "ffff80100000000100000000000000050000000e", // 16, 1, 0, 5, 14
+            "ffff801100000001000000000000000500000010", // 17, 1, 0, 5, 16
+        ];
         self.pack_u32(OP_CONNECT);
         self.pack_u32(OP_ATTACH);
         self.pack_u32(3); // CONNECT_VERSION3
         self.pack_u32(1); // Arch Type(GENERIC)
         self.pack_str(db_name);
-        self.pack_u32(1); // protocol count
+        self.pack_u32(protocols.len() as u32); // protocol count
         self.pack_bytes(&self.uid(
             &username,
             password,
@@ -845,8 +877,10 @@ impl WireProtocol {
             options["wire_crypt"] == "true",
             client_public,
         ));
-        // PROTOCOL_VERSION, Arch type (Generic=1), min, max, weight = 13, 1, 0, 5, 8
-        self.append_bytes(&hex::decode("ffff800d00000001000000000000000500000008").unwrap());
+
+        for p in protocols.iter() {
+            self.append_bytes(&hex::decode(p).unwrap());
+        }
         self.send_packets()?;
 
         Ok(())
@@ -975,10 +1009,10 @@ impl WireProtocol {
         Ok(())
     }
 
-    pub fn op_crypt(&mut self) -> Result<(), Error> {
+    pub fn op_crypt(&mut self, plugin: &Vec<u8>) -> Result<(), Error> {
         debug_print!("op_crypt()");
         self.pack_u32(OP_CRYPT);
-        self.pack_str("Arc4");
+        self.pack_bytes(plugin);
         self.pack_str("Symmetric");
         self.send_packets()?;
 
@@ -1157,6 +1191,11 @@ impl WireProtocol {
             self.pack_u32(1);
             self.append_bytes(&values);
         }
+        if self.protocol_version >= 16 {
+            // statement timeout
+            self.append_bytes(&vec![0; 4]);
+        }
+
         self.send_packets()?;
         Ok(())
     }
@@ -1186,6 +1225,10 @@ impl WireProtocol {
         }
         self.pack_bytes(output_blr);
         self.pack_u32(0);
+        if self.protocol_version >= 16 {
+            // statement timeout
+            self.append_bytes(&vec![0; 4]);
+        }
 
         self.send_packets()?;
         Ok(())
